/*
Copyright 2011 Johan Maasing

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
 */
package nu.zoom.catonine.swing.tail.gutter;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Graphics;
import java.awt.Point;
import java.awt.Rectangle;
import java.util.LinkedList;
import java.util.List;

import javax.swing.JComponent;
import javax.swing.JTextPane;
import javax.swing.text.BadLocationException;

import nu.zoom.catonine.swing.tail.TailingDocument.DocumentLogEntry;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

/**
 * @author "Johan Maasing" &lt;johan@zoom.nu&gt;
 * 
 */
public class GutterView extends JComponent {
	private static final long serialVersionUID = 6189389750396872064L;
	private final static int HEIGHT = Integer.MAX_VALUE - 1000000;
	private static final int GUTTER_Y_MARGIN = 4;
	private final Log log = LogFactory.getLog(getClass());
	private final GutterDocument document;
	private final JTextPane textArea;
	private CachedValues cv;

	public GutterView(final JTextPane textArea, final GutterDocument document) {
		super();
		this.document = document;
		this.textArea = textArea;
	}

	@Override
	protected void paintComponent(final Graphics gr) {
		super.paintComponent(gr);
		updateCachedValues();

		Graphics g = gr.create();
		final Rectangle clipBounds = g.getClipBounds();
		g.setFont(this.cv.gutterFont);
		FontMetrics fontMetrics = this.getFontMetrics(this.cv.gutterFont);

		try {
			List<OffsetCollection> offsets = calculateOffsets(clipBounds);
			for (OffsetCollection offset : offsets) {
				String text = Long.toString(offset.filePosition);
				final int stringXOffset = fontMetrics.stringWidth(text);
				g.setColor(Color.BLACK);
				g.drawLine(clipBounds.x, clipBounds.y + offset.y, clipBounds.x + clipBounds.width, clipBounds.y
						+ offset.y);
				g.drawString(text, clipBounds.x - GUTTER_Y_MARGIN + this.cv.gutterWidth - stringXOffset, clipBounds.y
						+ offset.y + this.cv.baseline);
			}
		} catch (BadLocationException e) {
			log.error(e);
		}
	}

	private List<OffsetCollection> calculateOffsets(final Rectangle clipBounds) throws BadLocationException {
		final LinkedList<OffsetCollection> offsets = new LinkedList<GutterView.OffsetCollection>();

		final Point viewStartPoint = new Point(0, clipBounds.y);
		final Point viewEndPoint = new Point(clipBounds.width, clipBounds.y + clipBounds.height);
		int modelStartOffset = this.textArea.viewToModel(viewStartPoint);
		int modelEndOffset = this.textArea.viewToModel(viewEndPoint);

		final List<DocumentLogEntry> intersectingLogEntries =
				this.document.getIntersectingLogEntries(modelStartOffset, modelEndOffset);

		for (final DocumentLogEntry documentLogEntry : intersectingLogEntries) {
			final Rectangle entryViewPosition = this.textArea.modelToView(documentLogEntry.offset);
			final OffsetCollection oc = new OffsetCollection();
			oc.filePosition = documentLogEntry.entryFileStart;
			oc.y = entryViewPosition.y - viewStartPoint.y;
			OffsetCollection last = offsets.peekLast();
			if (last != null) {
				last.height = oc.y - last.y;
			}
			offsets.add(oc);
		}
		OffsetCollection last = offsets.peekLast();
		if (last != null) {
			last.height = clipBounds.height - last.y;
		}

		return offsets;
	}

	static class OffsetCollection {
		long filePosition;
		int height;
		int y;
	}

	private void updateCachedValues() {
		if (this.cv == null) {
			CachedValues cv = new CachedValues();
			FontMetrics fontMetrics = textArea.getFontMetrics(textArea.getFont());
			cv.fontHeight = fontMetrics.getHeight();
			cv.baseline = cv.fontHeight - fontMetrics.getDescent();
			cv.gutterFont = new Font(Font.SANS_SERIF, Font.PLAIN, cv.fontHeight - 6);

			FontMetrics gutterFontMetrics = getFontMetrics(cv.gutterFont);
			cv.gutterWidth = gutterFontMetrics.charWidth('0') * 10 + 2 * GUTTER_Y_MARGIN;

			Dimension d = new Dimension(cv.gutterWidth, HEIGHT);
			setPreferredSize(d);
			setSize(d);

			this.cv = cv;
		}
	}

	static class CachedValues {
		Font gutterFont;
		int baseline;
		int lastHeight;
		int gutterWidth;
		int fontHeight;
	}
}
